;;; dml-defs.el --- Various definitions for dml-mode

;; Copyright (C) 2012  Ian Zerny <ian@zerny.dk>
;; Copyright (C) 1999, 2000, 2004  Stefan Monnier <monnier@gnu.org>
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:


;;; Code:

(eval-when-compile (require 'cl))
(require 'dml-util)


(defgroup dml ()
  "Editing DML code."
  :group 'languages)

(defvar dml-outline-regexp
  ;; `st' and `si' are to match structure and signature.
  "\\|s[ti]\\|[ \t]*\\(let[ \t]+\\)?\\(fun\\|and\\)\\>"
  "Regexp matching a major heading.
This actually can't work without extending `outline-minor-mode' with the
notion of \"the end of an outline\".")

;;; 
;;; Internal defines
;;; 

(defmap dml-mode-map
  ;; smarter cursor movement
  '((forward-sexp	. dml-user-forward-sexp)
    (backward-sexp	. dml-user-backward-sexp)
    ;; Text-formatting commands:
    ("\C-c\C-m"	. dml-insert-form)
    ("\C-c\C-i"	. dml-mode-info)
    ("\M-|"	. dml-electric-pipe)
    ("\M-\ "	. dml-electric-space)
    ("\;"	. dml-electric-semi)
    ("\M-\t"	. dml-back-to-outer-indent)
    ;; Process commands added to dml-mode-map -- these should autoload
    ("\C-c\C-l"	. dml-load-file)
    ("\C-c\C-c" . dml-compile)
    ("\C-c\C-s" . switch-to-dml)
    ("\C-c\C-r" . dml-send-region)
    ("\C-c\C-b" . dml-send-buffer)
    ([(meta shift down-mouse-1)] . dml-drag-region))
  "The keymap used in `dml-mode'."
  ;; :inherit dml-bindings
  :group 'dml)

(defsyntax dml-mode-syntax-table
  `((?\*   . ,(if dml-builtin-nested-comments-flag ". 23n" ". 23"))
    (?\(   . "()1")
    (?\)   . ")(4")
    ("._'" . "_")
    (",;"  . ".")
    ;; `!' is not really a prefix-char, oh well!
    ("~#!" . "'")
    ("%&$+-/:<=>?@`^|"	 . "."))
  "The syntax table used in `dml-mode'.")


(easy-menu-define dml-mode-menu dml-mode-map "Menu used in `dml-mode'."
  '("DML"
    ("Process"
     ["Start default ML compiler" run-dml		t]
     ["-" nil nil]
     ["run CM.make"		dml-compile	t]
     ["load ML source file"	dml-load-file	t]
     ["switch to ML buffer"	switch-to-dml	t]
     ["--" nil nil]
     ["send buffer contents"	dml-send-buffer	t]
     ["send region"		dml-send-region	t]
     ["send paragraph"		dml-send-function t]
     ["goto next error"		next-error	(featurep 'dml-proc)]
     ["---" nil nil]
     ;; ["Standard ML of New Jersey" dml-dmlnj	(fboundp 'dml-dmlnj)]
     ;; ["Poly/ML"			dml-poly-ml	(fboundp 'dml-poly-ml)]
     ;; ["Moscow ML"		dml-modml	(fboundp 'dml-modml)]
     ["Help for Inferior ML"	(describe-function 'inferior-dml-mode) :active (featurep 'dml-proc)])
    ["electric pipe"     dml-electric-pipe t]
    ["insert DML form"   dml-insert-form t]
    ("Forms" :filter dml-forms-menu)
    ("Format/Mode Variables"
     ["indent region"             indent-region t]
     ["outdent"                   dml-back-to-outer-indent t]
     ["-" nil nil]
     ["set indent-level"          dml-indent-level t]
     ["set pipe-indent"           dml-pipe-indent t]
     ["--" nil nil]
     ["toggle type-of-indent"     (dml-type-of-indent) t]
     ["toggle nested-if-indent"   (dml-nested-if-indent) t]
     ["toggle case-indent"        (dml-case-indent) t]
     ["toggle electric-semi-mode" (dml-electric-semi-mode) t])
    ["-----" nil nil]
    ["DML mode help (brief)"       describe-mode t]
    ["DML mode *info*"             dml-mode-info t]
    ["-----" nil nil]
    ["Remove overlay"    (dml-error-overlay 'undo) t ;:active (dml-overlay-active-p)
     ]))

;; Make's sure they appear in the menu bar when dml-mode-map is active.
;; On the hook for XEmacs only -- see easy-menu-add in auc-menu.el.
;; (defun dml-mode-menu-bar ()
;;   "Make sure menus appear in the menu bar as well as under mouse 3."
;;   (and (eq major-mode 'dml-mode)
;;        (easy-menu-add dml-mode-menu dml-mode-map)))
;; (add-hook 'dml-mode-hook 'dml-mode-menu-bar)

;;
;; regexps
;;

(defun dml-syms-re (&rest syms)
  (concat "\\<" (regexp-opt (flatten syms) t) "\\>"))

;;

(defconst dml-module-head-syms
  '("signature" "structure" "functor" "abstraction"))


(defconst dml-begin-syms
  '("let" "abstype" "local" "struct" "sig")
  "Symbols matching the `end' symbol.")

(defconst dml-begin-syms-re
  (dml-syms-re dml-begin-syms)
  "Symbols matching the `end' symbol.")

;; (defconst dml-user-begin-symbols-re
;;   (dml-syms-re "let" "abstype" "local" "struct" "sig" "in" "with")
;;   "Symbols matching (loosely) the `end' symbol.")

(defconst dml-sexp-head-symbols-re
  (dml-syms-re "let" "abstype" "local" "struct" "sig" "in" "with"
	       "if" "then" "else" "case" "of" "fn" "fun" "val" "and"
	       "datatype" "type" "exception" "open" "infix" "infixr" "nonfix"
	       "IF" "THEN" "ELSE" "FUN" "FN"
	       dml-module-head-syms
	       "handle" "raise")
  "Symbols starting an sexp.")

;; (defconst dml-not-arg-start-re
;;   (dml-syms-re "in" "of" "end" "andalso")
;;   "Symbols that can't be found at the head of an arg.")

;; (defconst dml-not-arg-re
;;   (dml-syms-re "in" "of" "end" "andalso")
;;   "Symbols that should not be confused with an arg.")

(defconst dml-=-starter-syms
  (list* "|" "val" "fun" "and" "datatype" "type" "abstype" "eqtype"
	 dml-module-head-syms)
  "Symbols that can be followed by a `='.")
(defconst dml-=-starter-re
  (concat "\\S.|\\S.\\|" (dml-syms-re (cdr dml-=-starter-syms)))
  "Symbols that can be followed by a `='.")

(defconst dml-indent-rule
  (dml-preproc-alist
   `(("struct" . 0)
     (,dml-module-head-syms "d=" 0)
     ("local" "in" 0)
     ;;("of" . (3 nil))
     ;;("else" . (dml-indent-level 0))
     ;;(("in" "fun" "and" "of") . (dml-indent-level nil))
     ("if" "else" 0)
     ("IF" "ELSE" 0)
     (,dml-=-starter-syms nil)
     (("abstype" "case" "datatype" "if" "then" "else" "sharing" "infix" "infixr"
       "let" "local" "nonfix" "open" "raise" "sig" "struct" "type" "val" "while"
       "do" "with" "withtype"
       "IF" "THEN" "ELSE")))))

(defconst dml-starters-indent-after
  (dml-syms-re "let" "local" "struct" "in" "sig" "with")
  "Indent after these.")

(defconst dml-delegate
  (dml-preproc-alist
   `((("of" "else" "then" "ELSE" "THEN" "d=") . (not (dml-bolp)))
     ("in" . t)))
  "Words which might delegate indentation to their parent.")

(defcustom dml-symbol-indent
  '(("fn" . -3)
    ("of" . 1)
    ("|" . -2)
    ("," . -2)
    (";" . -2)
    ;;("in" . 1)
    ("d=" . 2))
  "Special indentation alist for some symbols.
An entry like (\"in\" . 1) indicates that a line starting with the
symbol `in' should be indented one char further to the right.
This is only used in a few specific cases, so it does not work
for all symbols and in all lines starting with the given symbol."
  :group 'dml
  :type '(repeat (cons string integer)))

(defconst dml-open-paren
  (dml-preproc-alist
   `((,(list* "with" "in" dml-begin-syms) ,dml-begin-syms-re "\\<end\\>")))
  "Symbols that should behave somewhat like opening parens.")

(defconst dml-close-paren
  `(("in" "\\<l\\(ocal\\|et\\)\\>")
    ("with" "\\<abstype\\>")
    ("withtype" "\\<\\(abs\\|data\\)type\\>")
    ("end" ,dml-begin-syms-re)
    ("then" "\\<if\\>")
    ("else" "\\<if\\>" (dml-bolp))
    ("THEN" "\\<IF\\>")
    ("ELSE" "\\<IF\\>" (dml-bolp))
    ("of" "\\<case\\>")
    ("d=" nil))
  "Symbols that should behave somewhat like close parens.")

(defconst dml-agglomerate-re "\\<else[ \t]+if\\>"
  "Regexp of compound symbols (pairs of symbols to be considered as one).")

(defconst dml-non-nested-of-starter-re
  (dml-syms-re "datatype" "abstype" "exception")
  "Symbols that can introduce an `of' that shouldn't behave like a paren.")

(defconst dml-starters-syms
  (append dml-module-head-syms
	  '("abstype" "datatype" "exception" "fun"
	    "local" "infix" "infixr" "sharing" "nonfix"
	    "open" "type" "val" "and"
	    "withtype" "with"))
  "The starters of new expressions.")

(defconst dml-exptrail-syms
  '("if" "then" "else" "while" "withtype" "do" "case" "of" "raise" "fn" "IF" "THEN" "ELSE" "FN"))

(defconst dml-pipeheads
   '("|" "of" "fun" "fn" "and" "handle" "datatype" "abstype" "FUN" "FN")
   "A `|' corresponds to one of these.")


(provide 'dml-defs)

;;; dml-defs.el ends here
